﻿
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.Windows.Forms;
using System.IO.Ports;
using System.IO;
using EZ_Builder;
using EZ_Builder.Config;
using EZ_Builder.UCForms;

namespace XV_11
{
    /// <summary>
    /// Uses the EZ_Builder.UCForms.FormFluginMaster object so that this can be setup as an EZ-Builder plugin for others to use.
    /// Also initializes a lot of global variables that could be moved down to be local variables probably.  
    /// </summary>
    public partial class MainForm : EZ_Builder.UCForms.FormPluginMaster
    {
        /// <summary>
        /// variable declarations
        /// </summary>
        int imagecounter = 0;
        private SerialPort _serialPort = new SerialPort();
        private int _baudRate = 115200;
        private int _dataBits = 8;
        private Handshake _handshake = Handshake.None;
        private Parity _parity = Parity.None;
        private StopBits _stopBits = StopBits.One;
        private string tString = string.Empty;
        private string workingString = "";
        delegate void SetPanelCallback();
        private int distance = 0;
        private Int32 direction = 0;
        private int intensity = 0;
        List<int> lidarData = new List<int>();
        Graphics drawArea;
        delegate void SetTextCallback(string text);
        string minangle = "";
        string maxangle = "";
        string rotationangle = "";
        bool comportInitialized = false;
        bool drawline = true;

        /// <summary>
        /// self explanitory.  Comment not needed.  Feel stupid for even writing this comment, but for the sake of standards and all...
        /// </summary>
        public MainForm()
        {
            InitializeComponent();
        }

        /// <summary>
        /// sets up the COM port for use and opens it.
        /// </summary>
        private void initializecomport()
        {
            try {
                _serialPort.BaudRate = _baudRate;
                _serialPort.DataBits = _dataBits;
                _serialPort.Handshake = _handshake;
                _serialPort.Parity = _parity;
                _serialPort.PortName = (string)_cf.STORAGE[ConfigDictionary.comportname];
                _serialPort.StopBits = _stopBits;
                _serialPort.ReadBufferSize = Convert.ToInt32(_cf.STORAGE[ConfigDictionary.BufferSizeValue]);
                _serialPort.DataReceived += new SerialDataReceivedEventHandler(_serialPort_DataReceived);
                _serialPort.Open();
                comportInitialized = true;
                //_cf = cf;
            }
            catch(Exception ex)
            {
                MessageBox.Show(ex.Message.ToString());
                MessageBox.Show("COM port not initialized");
            }
        }
        
        /// <summary>
        /// function to open the comport and set the lidar to the proper mode of operation
        /// </summary>
        /// <returns></returns>
        public bool Open()
        {
            if (comportInitialized)
            {
                try
                {
                    //_serialPort.PortName = Convert.ToString(_cf.STORAGE[ConfigDictionary.comportname]);
                    _serialPort.WriteLine("ShowDist");
                    Thread.Sleep(50);
                    _serialPort.WriteLine("MotorOn");
                    Thread.Sleep(50);
                }
                catch(Exception ex)
                {
                    MessageBox.Show(ex.Message.ToString());
                    return false;
                }
            }
            else
            {
                initializecomport();
                Thread.Sleep(50);
                try
                {
                    _serialPort.PortName = Convert.ToString(_cf.STORAGE[ConfigDictionary.comportname]);
                    _serialPort.WriteLine("ShowDist");
                    Thread.Sleep(50);
                    _serialPort.WriteLine("MotorOn");
                    Thread.Sleep(50);
                }
                catch { return false; }
            }
            return true;

        }
        
        /// <summary>
        /// sets the necessary variables for use later when generating the map and then gets the values 
        /// from the lidar if the port is open
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button1_Click(object sender, EventArgs e)
        {
            
            drawArea = panel1.CreateGraphics();
            if (button1.Text == "Start LIDAR")
            {
                minangle = (string)_cf.STORAGE[ConfigDictionary.StartAngle];
                maxangle = (string)_cf.STORAGE[ConfigDictionary.StopAngle];
                rotationangle = (string)_cf.STORAGE[ConfigDictionary.ImageRotationValue];

                if (Open())
                {
                    int i = 0;
                    do
                    {
                        _serialPort_DataReceived(null, null);
                        _serialPort.DiscardInBuffer();
                        i++;
                        Thread.Sleep(50);
                    } while (i < 1);
                    
                    EZ_Builder.Invokers.SetText(button1, "Stop LIDAR");
                    EZ_Builder.Invokers.SetText(label1, "The LIDAR has been started");
                }
            }
            else
            {
                button1.Text = "Start LIDAR";
                //stoplidar = true;
                if (shutdownLidar())
                {
                    try
                    {
                        label1.Text = "The LIDAR has been stopped";
                        string pc = Convert.ToString(_cf.STORAGE[ConfigDictionary.thePenColor]);
                        string bc = Convert.ToString(_cf.STORAGE[ConfigDictionary.backgroundcolor]);
                        Pen pointpen = new Pen(Color.FromName(pc));
                        drawArea.Clear(Color.FromName(bc));
                        button1.Text = "Start LIDAR";
                        drawArea.DrawRectangle(pointpen, panel1.Width / 2, panel1.Height / 2, 4, 4);
                        imagecounter = 0;
                        drawArea.Dispose();
                    }
                    catch (Exception ex)
                    {
                        //MessageBox.Show(ex.Message.ToString());
                    }
                }
                else
                {
                    try
                    {
                        MessageBox.Show("I was unable to stop the LIDAR Successfully.  I will attempt to stop it again.");
                        shutdownLidar();
                        label1.Text = "The LIDAR has been stopped";
                        string pc = Convert.ToString(_cf.STORAGE[ConfigDictionary.thePenColor]);
                        string bc = Convert.ToString(_cf.STORAGE[ConfigDictionary.backgroundcolor]);
                        Pen pointpen = new Pen(Color.FromName(pc));
                        drawArea.Clear(Color.FromName(bc));
                        button1.Text = "Start LIDAR";
                        drawArea.DrawRectangle(pointpen, panel1.Width / 2, panel1.Height / 2, 4, 4);
                        imagecounter = 0;
                        drawArea.Dispose();
                    }
                    catch
                    {

                    }
                }
            }
            
        }
        
        /// <summary>
        /// stops the LIDAR motor and stops gathering data from the lidar.
        /// </summary>
        private bool shutdownLidar()
        {
            try
            {
                _serialPort.WriteLine("MotorOff");
                Thread.Sleep(50);
               // _serialPort.Dispose();
            }
            catch { return false; }
            return true;
        }

        /// <summary>
        /// reads the distances and angles from the serial feed from the LIDAR.
        /// Also, if the createfile variable is set to true at the top of this class
        /// it generates a text file showing what the feed was from the lidar every time
        /// this process runs.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void _serialPort_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            //Thread.Sleep(500);
            try {
                _serialPort.DiscardInBuffer();
                tString = "";
                Thread.Sleep(250);
                byte[] buffer = new byte[_serialPort.ReadBufferSize];
                int bytesRead = _serialPort.Read(buffer, 0, buffer.Length);
                tString += Encoding.ASCII.GetString(buffer, 0, bytesRead);
                workingString = tString;
                
                try
                {
                    BuildLidarArray(workingString);
                }
                catch
                {
                }
            }
            //catch(Exception ex)
            //{
                //MessageBox.Show(ex.Message.ToString());
            //}
            catch
            {
                //do noting
            }
        }

        /// <summary>
        /// bad name for this because the purpose of this has changed...
        /// takes the point from the lidar feed and gets the distance and angle reported
        /// back from the lidar.  There is cleanup to do in here for sure...
        /// </summary>
        /// <param name="data"></param>
        private void BuildLidarArray(string data)
        {
            string[] stringSeparators = new string[] { "\r\n" };
            string[] lines = data.Split(stringSeparators, StringSplitOptions.None);
            List<string> LidarData = lines.ToList<string>();
            if (LidarData.Count > 1)
            {
                LidarData.Sort();

                int counter = 1;
                foreach (string point in LidarData)
                {
                    if (counter > 0 && counter < 361)
                    {
                        string pointstring = point;
                        pointstring = pointstring.Replace(" ", "");
                        pointstring = pointstring.Replace(")", "");
                        pointstring = pointstring.Replace("r", "");
                        pointstring = pointstring.Replace("n", "");
                        pointstring = pointstring.Replace(@"\", "");
                        pointstring = pointstring.Replace(":", "|");
                        pointstring = pointstring.Replace("(", "|");

                        try
                        {
                            int pipepossition = pointstring.IndexOf("|");
                            int secondpipepossition = pointstring.IndexOf("|", pipepossition + 1);
                            int thirdpipepossition = pointstring.IndexOf("|", secondpipepossition + 1);
                            int readlength = (secondpipepossition - pipepossition - 1);

                            direction = Int32.Parse(pointstring.Substring(0, pipepossition));
                            distance = int.Parse(pointstring.Substring(pipepossition + 1, readlength));
                            intensity = int.Parse(pointstring.Substring(secondpipepossition + 1, pointstring.Length - secondpipepossition - 1));
                            if (Convert.ToBoolean(_cf.STORAGE[ConfigDictionary.gtChecked]))
                            {
                                if (direction > Convert.ToInt16(_cf.STORAGE[ConfigDictionary.StartAngle]) - 1 && direction < Convert.ToInt16(_cf.STORAGE[ConfigDictionary.StopAngle]) + 1)
                                {
                                    if (distance > 0)
                                    {

                                        string text = "Direction = " + direction.ToString() + " Distance = " + distance.ToString() + "Intensity = " + intensity.ToString() + Environment.NewLine;
                                        MapDraw(text);
                                        string distancetext = distance.ToString();
                                        Int32 direction2 = Int32.Parse(direction.ToString());

                                        EZ_Builder.Scripting.VariableManager.VariablesArray["$LIDAR"].ElementAt(direction2).SetValue(distancetext + "|" + intensity.ToString());
                                    }
                                    else
                                    {
                                        Int32 direction2 = Int32.Parse(direction.ToString());
                                        EZ_Builder.Scripting.VariableManager.VariablesArray["$LIDAR"].ElementAt(direction2).SetValue("0|0");
                                    }
                                }

                            }
                            if (!(bool)_cf.STORAGE[ConfigDictionary.gtChecked])
                            {
                                if (direction < int.Parse(minangle) + 1)
                                {
                                    if (distance > 0)
                                    {
                                        string text = "Direction = " + direction.ToString() + " Distance = " + distance.ToString() + Environment.NewLine; ;
                                        MapDraw(text);
                                        string distancetext = distance.ToString();
                                        Int32 direction2 = Int32.Parse(direction.ToString());

                                        EZ_Builder.Scripting.VariableManager.VariablesArray["$LIDAR"].ElementAt(direction2).SetValue(distancetext + "|" + intensity.ToString());
                                    }
                                    else
                                    {
                                        Int32 direction2 = Int32.Parse(direction.ToString());
                                        EZ_Builder.Scripting.VariableManager.VariablesArray["$LIDAR"].ElementAt(direction2).SetValue("0|0");
                                    }
                                }
                                if (direction > int.Parse(maxangle) - 1)
                                {
                                    if (distance > 0)
                                    {
                                        string text = "Direction = " + direction.ToString() + " Distance = " + distance.ToString() + Environment.NewLine; ;
                                        MapDraw(text);
                                        string distancetext = distance.ToString();
                                        Int32 direction2 = Int32.Parse(direction.ToString());

                                        EZ_Builder.Scripting.VariableManager.VariablesArray["$LIDAR"].ElementAt(direction2).SetValue(distancetext + "|" + intensity.ToString());
                                    }
                                    else
                                    {
                                        Int32 direction2 = Int32.Parse(direction.ToString());
                                        EZ_Builder.Scripting.VariableManager.VariablesArray["$LIDAR"].ElementAt(direction2).SetValue("0|0");
                                    }
                                }
                            }
                        }
                        catch
                        { }
                        //catch (Exception ex)
                        //{
                            //MessageBox.Show(ex.Message.ToString());
                        //}
                    }
                    else
                    {
                        counter = 0;
                    }

                    counter++;
                }
            }

       
           
        }
        
        /// <summary>
        /// used to plot points on panel1.  
        /// </summary>
        /// <param name="text"></param>
        private void MapDraw(string text)
        {

            double angle_rad = (direction * Math.PI / 180) + double.Parse(rotationangle);
            double c = Math.Cos(angle_rad);
            double s = -1 * Math.Sin(angle_rad);

            int dist_mm = distance;
            float dist_x = Convert.ToSingle((dist_mm * c) / Convert.ToInt32(_cf.STORAGE[ConfigDictionary.mapzoom]));
            float dist_y = Convert.ToSingle((dist_mm * s) / Convert.ToInt32(_cf.STORAGE[ConfigDictionary.mapzoom]));

            //drawArea = panel1.CreateGraphics();

            string pc = Convert.ToString(_cf.STORAGE[ConfigDictionary.thePen2Color]);
            string lc = Convert.ToString(_cf.STORAGE[ConfigDictionary.LineColorName]); 
            string bc = Convert.ToString(_cf.STORAGE[ConfigDictionary.backgroundcolor]);
            
            if(intensity < Convert.ToInt32(_cf.STORAGE[ConfigDictionary.thePenI1Color]))
            {
                pc = Convert.ToString(_cf.STORAGE[ConfigDictionary.thePenColor]);
            }
            if (intensity > Convert.ToInt32(_cf.STORAGE[ConfigDictionary.thePenI2Color]))
            {
                pc = Convert.ToString(_cf.STORAGE[ConfigDictionary.thePen2Color]);
            }
            if (intensity > Convert.ToInt32(_cf.STORAGE[ConfigDictionary.thePenI3Color]))
            {
                pc = Convert.ToString(_cf.STORAGE[ConfigDictionary.thePen3Color]);
            }

            Pen pointpen = new Pen(Color.FromName(pc));
            Pen linepen = new Pen(Color.FromName(lc));

            if (Convert.ToBoolean(_cf.STORAGE[ConfigDictionary.gtChecked]))
            {
                if (imagecounter < Convert.ToInt32(_cf.STORAGE[ConfigDictionary.StopAngle]) - Convert.ToInt32(_cf.STORAGE[ConfigDictionary.StartAngle]))
                {
                    drawArea.DrawRectangle(pointpen, dist_x + panel1.Width / 2, dist_y + panel1.Height / 2, 2, 2);
                    imagecounter++;
                }
                else
                {
                    drawArea.Clear(Color.FromName(bc));
                    drawArea.DrawRectangle(pointpen, panel1.Width / 2, panel1.Height / 2, 4, 4);
                    imagecounter = 0;
                    if (((string)_cf.STORAGE[ConfigDictionary.Var1]).Length > 0)
                    { updateText(); }
                }
            }
            if (drawline)
            {
                drawArea.DrawLine(linepen, panel1.Width / 2, panel1.Height / 2, dist_x + panel1.Width / 2, dist_y + panel1.Height / 2);
            }
            if (!Convert.ToBoolean(_cf.STORAGE[ConfigDictionary.gtChecked]))
            {
                if (imagecounter < Convert.ToInt32(_cf.STORAGE[ConfigDictionary.StartAngle]) + (360 - Convert.ToInt32(_cf.STORAGE[ConfigDictionary.StopAngle])))
                {
                    drawArea.DrawRectangle(pointpen, dist_x + panel1.Width / 2, dist_y + panel1.Height / 2, 2, 2);
                    imagecounter++;

                }
                else
                {
                    drawArea.Clear(Color.FromName(bc));
                    drawArea.DrawRectangle(pointpen, panel1.Width / 2, panel1.Height / 2, 4, 4);
                    imagecounter = 0;
                    if (Convert.ToString(_cf.STORAGE[ConfigDictionary.Var1]).Length > 0)
                    { updateText(); }
                }
            }
            if (drawline)
            {
                drawArea.DrawLine(linepen, panel1.Width / 2, panel1.Height / 2, dist_x + panel1.Width / 2, dist_y + panel1.Height/2);
            }
        }

        /// <summary>
        /// updates the text on the map for the variables specified.
        /// </summary>
        private void updateText()
        {
            int possitiony = 0;

            string fc = (string)_cf.STORAGE[ConfigDictionary.textColorName];

            var brush = new SolidBrush(Color.FromName(fc));

            //    graphics.DrawString(cmd.AllArguments, arialFont, brush, secondLocation);
            var font = new Font("Arial", 20, FontStyle.Regular, GraphicsUnit.Pixel);

            try
            {
                drawArea.DrawString(EZ_Builder.Scripting.VariableManager.GetVariable((string)_cf.STORAGE[ConfigDictionary.Var1]), font, brush, 0, possitiony);
            }
            catch
            {
                MessageBox.Show("I couldn't find the variable that you specified.");
            }
            
        }
        
        /// 
        /// This method is called by EZ-Builder when it requests the configuration for your  plugin. 
        /// EZ-Builder will request the configuration when the control is created and when the project is saved.
        /// The data set in this configuration will be serialized and saved in the EZ-Builder user's project.
        /// Custom data may be stored in the STORAGE dictionary.
        /// 
        public override EZ_Builder.Config.Sub.PluginV1 GetConfiguration()
        {

            return base.GetConfiguration();
        }
        
        /// 
        /// This method is called by EZ-Builder when another control sends a command to this control using the EZ-Script ControlCommand() function.
        /// The available ControlCommand() functions for this control should be returned in the GetSupportedControlCommandsMethodsForSlave() override.
        /// 
        public override void SendCommand(string windowCommand, params string[] values)
        {

            base.SendCommand(windowCommand, values);
        }

        /// 
        /// This method is called by EZ-Builder for the CHEAT SHEET to receive all available ControlCommand() syntax for this control.
        /// When the ControlCommand() is called for this function, see the SendCommand() override.
        /// 
        public override object[] GetSupportedControlCommands()
        {

            return base.GetSupportedControlCommands();
        }

        /// <summary>
        /// loads valid entries into the combo boxes, sets default starting values and 
        /// creates the $LIDAR variable array.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_Load(object sender, EventArgs e)
        {
            //button1.Enabled = false;
            
            loadConfig();

            EZ_Builder.Scripting.VariableManager.CreateVariableArray("$LIDAR", 0, 360);

            try {
                if (Convert.ToString(_cf.STORAGE[ConfigDictionary.comportname]).Length > 3)
                {
                    initializecomport();
                }
            }
            catch
            {
                MessageBox.Show("Select the COM port being used by the LIDAR on the config screen.");
            }
            backgroundWorker1_DoWork(null, null);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ucConfigurationButton1_Click(object sender, EventArgs e)
        {
            using (FormConfig formConfig = new FormConfig(_cf))
                if (formConfig.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                    _cf = formConfig.GetConfig;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="cf"></param>
        public override void SetConfiguration(EZ_Builder.Config.Sub.PluginV1 cf)
        {
            cf.STORAGE.AddIfNotExist(ConfigDictionary.ImageRotation,"90");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.ImageRotationValue, "0");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.StartAngle,"0");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.StopAngle,"359");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.BufferSize,"31");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.BufferSizeValue, "2048");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.comportname,"COM3");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.gtChecked, "true");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.mapzoom,"16");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.background,"164");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.backgroundcolor, "Cyan");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.thePen,"10");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.thePenColor, "Blue");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.thePen2,"12");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.thePen2Color, "Brown");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.thePen3,"8");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.thePen3Color, "Black");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.thePenI1,"0");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.thePenI1Color, "0");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.thePenI2,"49");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.thePenI2Color, "2450");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.thePenI3,"72");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.thePenI3Color, "3600");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.Var1, "");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.textcolor,"8");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.textColorName, "Black");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.DrawLines, "true");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.LineColor,"44");
            cf.STORAGE.AddIfNotExist(ConfigDictionary.LineColorName, "Ivory");
            base.SetConfiguration(cf);
        }

        /// <summary>
        /// Loads the config file to set the variables.
        /// </summary>
        private void loadConfig()
        {
            drawline = Convert.ToBoolean(_cf.STORAGE[ConfigDictionary.DrawLines]);
            drawArea = panel1.CreateGraphics();
            string pc = Convert.ToString(_cf.STORAGE[ConfigDictionary.thePenColor]);
            string bc = Convert.ToString(_cf.STORAGE[ConfigDictionary.backgroundcolor]);
            Pen pointpen = new Pen(Color.FromName(pc));
            drawArea.Clear(Color.FromName(bc));
            drawArea.DrawRectangle(pointpen, panel1.Width / 2, panel1.Height / 2, 4, 4);
            drawArea.Dispose();

        }

        private void backgroundWorker1_DoWork(object sender, DoWorkEventArgs e)
        {
            EZ_Builder.Scripting.VariableManager.CreateVariableArray("$LIDAR", 0, 360);
        }

    }
}
